from abc import ABC, abstractmethod
import pyspiel
import os
from datetime import datetime
import sys

class GoBot(ABC):
    def _redirect_logs(self):
        """
        Redirects stdout and stderr to a log file named `{name}_bot_{timestamp}.log`.
        Writes logs to the `/logs` directory.
        """
        os.makedirs("/logs", exist_ok=True)
        os.makedirs(f"/logs/{self.name}", exist_ok=True)
        timestamp = datetime.now().strftime("%Y%m%d_%H%M%S")
        log_filename = f"/logs/{self.name}/{timestamp}.log"
        log_file = open(log_filename, "a")

        sys.stdout = log_file
        sys.stderr = log_file

        print(f"Logging started for bot: {self.name}")
    
    def on_start(self, player_data):
        """Called at the start of each match"""
        self.match_data = {}
        self._redirect_logs()
        self.match_data['player'] = player_data['player']
        self.setup()

    def on_update(self, game_state: pyspiel.State):
        """Called after each move is made"""
        self.update(game_state)

    def get_action(self, game_state: pyspiel.State) -> int:
        """Main method to get the bot's next move"""
        return self.get_move(game_state)

    def setup(self):
        """Override to add custom match start logic"""
        pass

    def update(self, game_state: pyspiel.State, last_move: int):
        """Override to add custom move tracking logic"""
        pass

    @abstractmethod
    def get_move(self, game_state: pyspiel.State, time_remaining: int) -> int:
        """Select a move given the current game state"""
        pass
    
    def __repr__(self):
        return str(self)


# tag::scoring_imports[]
from __future__ import absolute_import
from collections import namedtuple

from agt_arena.game.dlgo_go.gotypes import Player, Point
# end::scoring_imports[]


# tag::scoring_territory[]
class Territory:
    def __init__(self, territory_map):  # <1>
        self.num_black_territory = 0
        self.num_white_territory = 0
        self.num_black_stones = 0
        self.num_white_stones = 0
        self.num_dame = 0
        self.dame_points = []
        for point, status in territory_map.items():  # <2>
            if status == Player.black:
                self.num_black_stones += 1
            elif status == Player.white:
                self.num_white_stones += 1
            elif status == 'territory_b':
                self.num_black_territory += 1
            elif status == 'territory_w':
                self.num_white_territory += 1
            elif status == 'dame':
                self.num_dame += 1
                self.dame_points.append(point)

# <1> A `territory_map` splits the board into stones, territory and neutral points (dame).
# <2> Depending on the status of a point, we increment the respective counter.
# end::scoring_territory[]


# tag::scoring_game_result[]
class GameResult(namedtuple('GameResult', 'b w komi')):
    @property
    def winner(self):
        if self.b > self.w + self.komi:
            return Player.black
        return Player.white

    @property
    def winning_margin(self):
        w = self.w + self.komi
        return abs(self.b - w)

    def __str__(self):
        w = self.w + self.komi
        if self.b > w:
            return 'B+%.1f' % (self.b - w,)
        return 'W+%.1f' % (w - self.b,)
# end::scoring_game_result[]


""" evaluate_territory:
Map a board into territory and dame.

Any points that are completely surrounded by a single color are
counted as territory; it makes no attempt to identify even
trivially dead groups.
"""


# tag::scoring_evaluate_territory[]
def evaluate_territory(board):

    status = {}
    for r in range(1, board.num_rows + 1):
        for c in range(1, board.num_cols + 1):
            p = Point(row=r, col=c)
            if p in status:  # <1>
                continue
            stone = board.get(p)
            if stone is not None:  # <2>
                status[p] = board.get(p)
            else:
                group, neighbors = _collect_region(p, board)
                if len(neighbors) == 1:  # <3>
                    neighbor_stone = neighbors.pop()
                    stone_str = 'b' if neighbor_stone == Player.black else 'w'
                    fill_with = 'territory_' + stone_str
                else:
                    fill_with = 'dame'  # <4>
                for pos in group:
                    status[pos] = fill_with
    return Territory(status)

# <1> Skip the point, if you already visited this as part of a different group.
# <2> If the point is a stone, add it as status.
# <3> If a point is completely surrounded by black or white stones, count it as territory.
# <4> Otherwise the point has to be a neutral point, so we add it to dame.
# end::scoring_evaluate_territory[]


""" _collect_region:

Find the contiguous section of a board containing a point. Also
identify all the boundary points.
"""


# tag::scoring_collect_region[]
def _collect_region(start_pos, board, visited=None):

    if visited is None:
        visited = {}
    if start_pos in visited:
        return [], set()
    all_points = [start_pos]
    all_borders = set()
    visited[start_pos] = True
    here = board.get(start_pos)
    deltas = [(-1, 0), (1, 0), (0, -1), (0, 1)]
    for delta_r, delta_c in deltas:
        next_p = Point(row=start_pos.row + delta_r, col=start_pos.col + delta_c)
        if not board.is_on_grid(next_p):
            continue
        neighbor = board.get(next_p)
        if neighbor == here:
            points, borders = _collect_region(next_p, board, visited)
            all_points += points
            all_borders |= borders
        else:
            all_borders.add(neighbor)
    return all_points, all_borders
# end::scoring_collect_region[]


# tag::scoring_compute_game_result[]
def compute_game_result(game_state):
    territory = evaluate_territory(game_state.board)
    return GameResult(
        territory.num_black_territory + territory.num_black_stones,
        territory.num_white_territory + territory.num_white_stones,
        komi=7.5)
# end::scoring_compute_game_result[]


from agt_arena.game.dlgo_go.board import Board, GameState, Move
from agt_arena.game.dlgo_go.scoring import compute_game_result
from agt_arena.game.dlgo_go.gotypes import Player, Point
import random

class GoGame:
    def __init__(self, board_size=19):
        self.board_size = board_size
        self.game_state = GameState.new_game(self.board_size)
        self.move_history = []

    def play_move(self, move):
        if self.game_state.is_valid_move(move):
            self.game_state = self.game_state.apply_move(move)
            self.move_history.append(move)
            return True
        return False

    def score_game(self):
        return compute_game_result(self.game_state)
        
    def run_match(self, players):
        """
        Run a match between two players.
        Args:
            players: Tuple of (black_player, white_player) where players are GoBot instances
        Returns:
            Match result dictionary
        """
        black_player, white_player = players
        
        # Setup players for the match
        black_player.setup(Player.black)
        white_player.setup(Player.white)
        
        current_player = black_player
        self.move_history = []
        self.game_state = GameState.new_game(self.board_size)
        
        last_move = None
        while not self.game_state.is_over():
            # Update both players with the current state
            black_player.update(self.game_state, last_move)
            white_player.update(self.game_state, last_move)
            
            # Get the current player's move
            move = current_player.get_action(self.game_state)
            
            if move.is_resign:
                return {
                    'winner': white_player if current_player == black_player else black_player,
                    'resigned': True,
                    'moves': len(self.move_history)
                }
                
            if self.play_move(move):
                last_move = move
                current_player = white_player if current_player == black_player else black_player
            
        result = self.score_game()
        return {
            'winner': black_player if result.winner == Player.black else white_player,
            'score': str(result),
            'moves': len(self.move_history)
        }

    
    def generate_sgf(self, match):
        """
        Generate SGF string for the game.
        Args:
            match: Tuple of (black_player, white_player)
        Returns:
            SGF format string
        """
        black_player, white_player = match
        moves_str = ""
        
        for i, move in enumerate(self.move_history):
            color = 'B' if i % 2 == 0 else 'W'
            player = black_player if i % 2 == 0 else white_player
            
            if move.is_pass:
                move_str = "[]"
                moves_str += f";{color}{move_str}C[{player} passes]"
            elif move.is_resign:
                moves_str += f";{color}[]C[{player} resigns]"
                break
            else:
                x = chr(ord('a') + move.point.col - 1)
                y = chr(ord('a') + move.point.row - 1)
                move_str = f"[{x}{y}]"
                moves_str += f";{color}{move_str}"
            
        if self.game_state.is_over():
            result = self.score_game()
            if any(m.is_resign for m in self.move_history):
                winner = white_player if result.winner == Player.white else black_player
                moves_str += f"C[Game ended by resignation. {winner} wins]"
            else:
                moves_str += f"C[Final score: {str(result)}]"
            
        return f"(;FF[4]CA[UTF-8]SZ[{self.board_size}]PB[{black_player}]PW[{white_player}]{moves_str})"



from agt_arena.matchmaker import BaseMatchmaker
import itertools

class RoundRobinMatchmaker(BaseMatchmaker):
    def __init__(self, match_size=2, order_matters=False):
        """
        match_size: Number of players per match.
        order_matters: If True, generates permutations; otherwise, combinations.
        """
        self.match_size = match_size
        self.order_matters = order_matters

    def generate_matches(self, players):
        """
        Generate matches using round-robin logic.
        """
        if self.order_matters:
            return list(itertools.permutations(players, self.match_size))
        else:
            return list(itertools.combinations(players, self.match_size))



from abc import ABC, abstractmethod

class BaseGame(ABC):
    """
    Abstract base class for all games.
    """

    @abstractmethod
    def run_match(self, players):
        """
        Run a single match and return detailed results.
        Args:
            players: List of players.
        Returns:
            dict: Match results, including placements.
        """
        pass



from .go_game import GoGame
from .board import Board, GameState, Move
from .gotypes import Player, Point
from .scoring import compute_game_result

__all__ = [
    "GoGame",
    "Board",
    "GameState",
    "Move",
    "Player",
    "Point",
]



import itertools
from agt_arena.matchmaker.base_matchmaker import BaseMatchmaker
import os 
import json


class EloMatchmaker(BaseMatchmaker):
    def __init__(self, n_closest=3, k_factor=32, order_matters=False, meta_file="meta/results.json"):
        super().__init__()
        self.n_closest = n_closest
        self.k_factor = k_factor
        self.order_matters = order_matters
        self.meta_file = meta_file
        
        self.load_meta_file()
        
    def load_meta_file(self):
        """Load ELO data from the meta file if it exists."""
        if os.path.exists(self.meta_file):
            try:
                with open(self.meta_file, "r") as file:
                    data = json.load(file)
                for bot, results in data.items():
                    elo = results['elo']
                    self.bot_data[bot] = {"elo": elo}
            except Exception as e:
                print(f"Error reading meta file {self.meta_file}: {e}")

    def update_elo(self, winner, loser):
        """Update ELO scores based on match results"""
        winner_elo = self.bot_data[winner]["elo"]
        loser_elo = self.bot_data[loser]["elo"]

        # Expected scores
        winner_expected = 1 / (1 + 10 ** ((loser_elo - winner_elo) / 400))
        loser_expected = 1 - winner_expected

        self.bot_data[winner]["elo"] += self.k_factor * (1 - winner_expected)
        self.bot_data[loser]["elo"] += self.k_factor * (0 - loser_expected)
        
    
    def get_elo(self, player): 
        return self.bot_data[player]["elo"]

    def add_new_bots(self, players):
        """Add bots to the matchmaker with default ELO if they don't already exist"""
        os.makedirs(os.path.dirname(self.meta_file), exist_ok=True)
        
        for bot in players:
            if bot not in self.bot_data:
                self.bot_data[bot] = {"elo": 1500}

    def generate_matches(self, players):
        """Generate matches for the closest n bots based on ELO (Generator)"""
        for bot in players:
            sorted_bots = sorted(
                [p for p in players if p != bot],
                key=lambda other: abs(self.bot_data[bot]["elo"] - self.bot_data[other]["elo"])
            )

            for opponent in sorted_bots[:self.n_closest]:
                yield (bot, opponent)
                if self.order_matters:
                    yield (opponent, bot)



import os
import logging
from typing import Dict, List
import requests
import time
from datetime import datetime
from agt_arena.matchmaker.elo import EloMatchmaker
from agt_arena.game.go.go_search_problem import GoState, GoProblem
import json
import docker
from tqdm import tqdm

class GoArena:
    def __init__(self, submission_ids: List[int], board_size, setupTimeout: int = 5, totalMoveTimeout: int = 15, grace_time_per_move: float = 0.05, port = 5000):
        """
        Args:
            submission_ids: List of unique id submissions
            board_size: Size of the Go board
            setupTimeout: Timeout in seconds for bot setup
            totalMoveTimeout: Total timeout in seconds for all moves
            grace_time_per_move: Additional grace time added to the total time after each move
        """
        
        self.bot_names = {}
        self.bot_urls = {}
        self.board_size = board_size
        self.setupTimeout = setupTimeout
        self.totalMoveTimeout = totalMoveTimeout
        self.grace_time_per_move = grace_time_per_move
        self.port = port
        self.health_checks(submission_ids)
        
        # Setup logging with timestamped filenames
        os.makedirs("logs", exist_ok=True)
        timestamp = datetime.now().strftime("%Y%m%d_%H%M%S")
        log_filename = f"logs/tournament_{timestamp}.log"
        logging.basicConfig(
            filename=log_filename,
            level=logging.INFO,
            format="%(asctime)s - %(levelname)s - %(message)s",
        )
        
        self.totals_path = "meta/totals.json"
        self.results_path = "meta/results.json"
    
    def health_checks(self, submission_ids): 
        for bot_id in tqdm(submission_ids, desc = "Running Health Checks"): 
            _, url = GoArena.build_and_run_container(bot_id, self.port, self.board_size)
            print(f"URL sucessfully built: {url}")
            try: 
                name = self.health_check(url)
                if name:
                    print(f"{name} has sucessfully joined to the arena")
                    self.bot_names[bot_id] = name
                    self.bot_urls[bot_id] = url
                    
                else:
                    print(f"Health check failed for bot-id: {bot_id}")
            finally: 
                GoArena.cleanup_container_and_image(bot_id)
            
    
    def health_check(self, url): 
        try:
            response = requests.get(f"{url}/health", timeout=15)
            if response.status_code == 200:
                status = response.json()
                if status.get("status") == "healthy" and "name" in status and isinstance(status["name"], str):
                    return status["name"]
            return None
        except requests.exceptions.RequestException as e:
            print(f"Health check failed for {url}: {e}")
            return None
        
        
    def build_and_run_container(bot_id, app_port, board_size, delay = 10):
        client = docker.from_env()
        image_name = f"bot_{bot_id}"
        container_name = f"bot_container_{bot_id}"
        bot_port = app_port  # Port for this bot

        # Build the Docker image
        print(f"Building Docker image for bot {bot_id}...")
        try:
            build_logs = client.api.build(
                path=os.path.abspath("./"),
                dockerfile=os.path.abspath(f"./submissions/go/submissions/{bot_id}/Dockerfile"),
                tag=image_name,
                rm=True,
                decode=True
            )
            for log in build_logs:
                if 'stream' in log:
                    print(log['stream'].strip())
                elif 'error' in log:
                    print(f"Build error: {log['error']}")
            print(f"Image for bot {bot_id} built successfully.")
        except docker.errors.BuildError as e:
            print(f"Error building image for bot {bot_id}: {e}")
            return None, None

        # Check if the image exists locally
        try:
            client.images.get(image_name)
        except docker.errors.ImageNotFound:
            print(f"Image {image_name} was not found after building. Exiting.")
            return None, None

        # Run the container
        print(f"Starting container for bot {bot_id}...")
        try:
            container = client.containers.run(
                image_name,
                name=container_name,
                detach=True,
                environment={
                    "BOT_PORT": bot_port,
                    "FLASK_ENV": "development",
                    "FLASK_DEBUG": "1",
                    "BOARD_SIZE": board_size,
                },
                healthcheck={
                    "test": ["CMD", "curl", "-f", f"http://{container_name}:{bot_port}/health"],
                    "interval": int(5e9),  # 5 seconds
                    "timeout": int(3e9),   # 3 seconds
                    "retries": 3,
                },
                volumes={
                    "/Users/johnwu/Documents/Brown CS/cs410_arena/logs/go": {"bind": "/logs", "mode": "rw"},
                },
                network="tournament",
                tty=True,
                stdin_open=True
            )
            print(f"Container for bot {bot_id} is running.")
            time.sleep(delay)
            return container, f"http://{container_name}:{bot_port}"
        except docker.errors.APIError as e:
            print(f"Error starting container for bot {bot_id}: {e}")
            return None, f"http://{container_name}:{bot_port}"


    def cleanup_container_and_image(bot_id):
        client = docker.from_env()
        container_name = f"bot_container_{bot_id}"
        image_name = f"bot_{bot_id}"

        # Stop and remove container
        print(f"Stopping and removing container for bot {bot_id}...")
        try:
            container = client.containers.get(container_name)
            container.stop()
            container.remove()
        except Exception as e:
            print(f"Error cleaning up container {container_name}: {e}")

        # # Remove image
        # print(f"Removing image for bot {bot_id}...")
        # try:
        #     client.images.remove(image_name, force=True)
        # except Exception as e:
        #     print(f"Error removing image {image_name}: {e}")
            
    def setup_bot(self, bot_url: str, player: int) -> bool:
        try:
            response = requests.post(
                f"{bot_url}/setup",
                json={'player': player},
                timeout=self.setupTimeout
            )
            return response.status_code == 200
        except requests.exceptions.RequestException as e:
            logging.error(f"Error setting up bot at {bot_url}: {e}")
            return False

    def get_bot_move(self, bot_url: str, game_state: GoState, time_remaining: float) -> int:
        try:
            start_time = time.time()
            print(game_state.serialize())
            response = requests.post(
                f"{bot_url}/get_move",
                json={
                    'game_state': game_state.serialize(),
                    'time_remaining': time_remaining 
                },
                timeout=min(self.totalMoveTimeout, time_remaining)
            )
            elapsed_time = time.time() - start_time
            move_data = response.json()['move']
            return move_data, elapsed_time
        except requests.exceptions.RequestException as e:
            logging.error(f"Error getting move from bot at {bot_url}: {e}")
            return -1, 0
        except Exception as e:
            logging.error(f"Bot at {bot_url} encountered an error: {e}")
            return -1, 0

    def save_sgf(self, moves, black_bot, white_bot):
        """Save the match to an SGF file."""
        try: 
            sgf_data = "(;GM[1]FF[4]SZ[{}]PB[{}]PW[{}]".format(self.board_size, black_bot, white_bot)
            count = 0
            for move in moves:
                row, col = divmod(move, self.board_size)
                sgf_move = f";B[{chr(97 + col)}{chr(97 + row)}]" if count % 2 == 1 else f";W[{chr(97 + col)}{chr(97 + row)}]"
                sgf_data += sgf_move
                count += 1
            sgf_data += ")"

            # Create directories for each bot
            black_dir = os.path.join("logs", black_bot)
            white_dir = os.path.join("logs", white_bot)
            os.makedirs(black_dir, exist_ok=True)
            os.makedirs(white_dir, exist_ok=True)

            # Save SGF files in each bot's directory
            match_timestamp = datetime.now().strftime("%Y%m%d_%H%M%S")
            black_sgf_path = os.path.join(black_dir, f"{match_timestamp}_vs_{white_bot}.sgf")
            white_sgf_path = os.path.join(white_dir, f"{match_timestamp}_vs_{black_bot}.sgf")

            with open(black_sgf_path, "w") as black_file, open(white_sgf_path, "w") as white_file:
                black_file.write(sgf_data)
                white_file.write(sgf_data)
        except:
            pass
    

    def run_match(self, black_bot: str, black_bot_id: int, white_bot: str, white_bot_id: int) -> dict:
        """Run a single match between two bots"""
        _, black_bot_url = GoArena.build_and_run_container(black_bot_id, self.port, self.board_size)
        _, white_bot_url = GoArena.build_and_run_container(white_bot_id, self.port, self.board_size)
        
        # black_bot_url = self.bot_urls[black_bot_id]
        # white_bot_url = self.bot_urls[white_bot_id]
        my_go = GoProblem(size=self.board_size)
        state = my_go.start_state
        remaining_time = {black_bot_url: self.totalMoveTimeout, white_bot_url: self.totalMoveTimeout}
        move_counts = 0

        moves = []
        
        try: 
            current_bot_url = black_bot_url
        
            while not my_go.is_terminal_state(state):
                move, elapsed_time = self.get_bot_move(current_bot_url, state, remaining_time[current_bot_url])
                print(move, elapsed_time)
                
                if move not in state.legal_actions() or elapsed_time > (remaining_time[current_bot_url] + self.grace_time_per_move):
                    logging.error(f"Bot at {current_bot_url} ran out of time or failed to provide a move.")
                    return {
                        'winner': 'white' if current_bot_url == black_bot_url else 'black',
                        'resigned': True,
                        'duration': elapsed_time, 
                        'move_counts': move_counts
                    }
                    
                moves.append(move)
                move_counts += 1

                # Update remaining time and add grace time
                remaining_time[current_bot_url] -= elapsed_time
                remaining_time[current_bot_url] = min(
                    remaining_time[current_bot_url] + self.grace_time_per_move, self.totalMoveTimeout
                )

                state = my_go.transition(state, move)
                current_bot_url = white_bot_url if current_bot_url == black_bot_url else black_bot_url

            result = state.terminal_value()
            
            self.save_sgf(moves, black_bot, white_bot)
            
            return {
                'winner': 'black' if result[0] > 0 else 'white',
                'score': str(result),
                'duration': elapsed_time,
                'move_counts': move_counts
            }
        finally:
            GoArena.cleanup_container_and_image(white_bot_id)
            GoArena.cleanup_container_and_image(black_bot_id)

    def run_tournament(self) -> List[dict]:
        """Run a round-robin tournament between all bots"""
        results = []
        bot_ids = list(self.bot_names.keys())
        
        try:
            # Load existing stats from totals_path
            with open(self.totals_path, "r") as f:
                existing_stats = json.load(f)
        except FileNotFoundError:
            # If totals_path doesn't exist, start fresh
            existing_stats = {}

        # Initialize bot stats, loading from existing stats if available, or using defaults otherwise
        bot_stats = {
            bot_id: existing_stats.get(
                bot_id, 
                {'wins': 0, 'losses': 0, 'total_score': 0.0, 'matches': 0, 'total_moves': 0, 'total_duration': 0.0}
            )
            for bot_id in bot_ids
        }
    
        matchmaker = EloMatchmaker(order_matters=True, meta_file=self.totals_path)
        matchmaker.add_new_bots(bot_ids)
        matchups = matchmaker.generate_matches(bot_ids)

    
        for black_bot_id, white_bot_id in tqdm(matchups, desc="Running matches"):
            black_bot = self.bot_names[black_bot_id]
            white_bot = self.bot_names[white_bot_id]
            print(f"\nStarting match: {black_bot} (B) vs {white_bot} (W)")
            result = self.run_match(black_bot, black_bot_id, white_bot, white_bot_id)

            if 'error' in result:
                logging.error(f"Error in match: {result['error']}")
                continue

            result.update({'black': black_bot, 'white': white_bot})
            results.append(result)
            logging.info(f"Match {black_bot} (B) vs {white_bot} (W) Result: {result}")

            # Update bot stats
            bot_stats[black_bot_id]['name'] = black_bot
            bot_stats[white_bot_id]['name'] = white_bot
            
            bot_stats[black_bot_id]['matches'] += 1
            bot_stats[white_bot_id]['matches'] += 1

            winner = result['winner']                 
            loser = black_bot_id if winner == 'white' else white_bot_id
            winner = black_bot_id if winner == 'black' else white_bot_id
            bot_stats[winner]['wins'] += 1
            bot_stats[loser]['losses'] += 1
            
            matchmaker.update_elo(winner = winner, loser = loser)
            bot_stats[black_bot_id]['elo'] = matchmaker.get_elo(black_bot_id)

            bot_stats[black_bot_id]['total_duration'] += result['duration']
            bot_stats[white_bot_id]['total_duration'] += result['duration']
            
            bot_stats[black_bot_id]['total_moves'] += result['move_counts']
            bot_stats[white_bot_id]['total_moves'] += result['move_counts']
            
            with open(self.totals_path, "w") as f:
                json.dump(bot_stats, f, indent=4)
        
        # Log detailed bot statistics
        results_data = {}
        for bot, stats in tqdm(bot_stats.items(), desc="Calculating Stats"):
            avg_score = stats['total_score'] / stats['matches'] if stats['matches'] > 0 else 0.0
            avg_moves = stats['total_moves'] / stats['matches'] if stats['matches'] > 0 else 0.0
            avg_dur = stats['total_duration'] / stats['matches'] if stats['matches'] > 0 else 0.0
            elo = matchmaker.get_elo(bot)
            logging.info(f"Bot {stats['name']}: Wins: {stats['wins']}, Losses: {stats['losses']}, Average Score: {avg_score}, Average Moves per Game: {avg_moves}, Average Duration: {avg_dur}, Elo: {elo}")
            results_data[bot] = {
                "name": stats['name'],
                "average_score": (stats["wins"] - stats['losses']) / stats["matches"] if stats["matches"] > 0 else 0.0,
                "average_moves": stats["total_moves"] / stats["matches"] if stats["matches"] > 0 else 0.0,
                "average_duration": stats["total_duration"] / stats["matches"] if stats["matches"] > 0 else 0.0,
                "elo": elo
            }
            
        timestamp = time.strftime("%Y%m%d_%H%M%S")
        results_path_timestamp = f"{self.results_path.rstrip('.json')}_{timestamp}.json"
        with open(results_path_timestamp, "w") as f:
            json.dump(results_data, f, indent=4)

        return results
